@@ -14,11 +14,12 @@ use yrs::types::map::{MapEvent, MapIter};
 use yrs::types::text::TextEvent;
 use yrs::types::xml::{Attributes, TreeWalker, XmlEvent, XmlTextEvent};
 use yrs::types::{
-    Attrs, Branch, BranchPtr, Change, Delta, EntryChange, Path, PathSegment, TypeRefs, Value,
-    TYPE_REFS_ARRAY, TYPE_REFS_MAP, TYPE_REFS_TEXT, TYPE_REFS_XML_ELEMENT, TYPE_REFS_XML_TEXT,
+    Attrs, Branch, BranchPtr, Change, DeepObservable, Delta, EntryChange, Event, Events, Path,
+    PathSegment, TypeRefs, Value, TYPE_REFS_ARRAY, TYPE_REFS_MAP, TYPE_REFS_TEXT,
+    TYPE_REFS_XML_ELEMENT, TYPE_REFS_XML_TEXT,
 };
-use yrs::updates::decoder::{Decode, DecoderV1};
-use yrs::updates::encoder::{Encode, Encoder, EncoderV1};
+use yrs::updates::decoder::{Decode, DecoderV1, DecoderV2};
+use yrs::updates::encoder::{Encode, Encoder, EncoderV1, EncoderV2};
 use yrs::{
     Array, Doc, Map, OffsetKind, Options, StateVector, Subscription, Text, Transaction, Update,
     Xml, XmlElement, XmlText,
@@ -237,6 +238,31 @@ pub fn encode_state_as_update(doc: &mut YDoc, vector: Option<Uint8Array>) -> Uin
     doc.begin_transaction().diff_v1(vector)
 }
 
+/// Encodes all updates that have happened since a given version `vector` into a compact delta
+/// representation using lib0 v2 encoding. If `vector` parameter has not been provided, generated
+/// delta payload will contain all changes of a current ywasm document, working effectivelly as its
+/// state snapshot.
+///
+/// Example:
+///
+/// ```javascript
+/// import {YDoc, encodeStateVector, encodeStateAsUpdate, applyUpdate} from 'ywasm'
+///
+/// /// document on machine A
+/// const localDoc = new YDoc()
+/// const localSV = encodeStateVector(localDoc)
+///
+/// // document on machine B
+/// const remoteDoc = new YDoc()
+/// const remoteDelta = encodeStateAsUpdateV2(remoteDoc, localSV)
+///
+/// applyUpdate(localDoc, remoteDelta)
+/// ```
+#[wasm_bindgen(js_name = encodeStateAsUpdateV2)]
+pub fn encode_state_as_update_v2(doc: &mut YDoc, vector: Option<Uint8Array>) -> Uint8Array {
+    doc.begin_transaction().diff_v2(vector)
+}
+
 /// Applies delta update generated by the remote document replica to a current document. This
 /// method assumes that a payload maintains lib0 v1 encoding format.
 ///
@@ -253,13 +279,36 @@ pub fn encode_state_as_update(doc: &mut YDoc, vector: Option<Uint8Array>) -> Uin
 /// const remoteDoc = new YDoc()
 /// const remoteDelta = encodeStateAsUpdate(remoteDoc, localSV)
 ///
-/// applyUpdate(localDoc, remoteDelta)
+/// applyUpdateV2(localDoc, remoteDelta)
 /// ```
 #[wasm_bindgen(js_name = applyUpdate)]
 pub fn apply_update(doc: &mut YDoc, diff: Uint8Array) {
     doc.begin_transaction().apply_v1(diff);
 }
 
+/// Applies delta update generated by the remote document replica to a current document. This
+/// method assumes that a payload maintains lib0 v2 encoding format.
+///
+/// Example:
+///
+/// ```javascript
+/// import {YDoc, encodeStateVector, encodeStateAsUpdate, applyUpdate} from 'ywasm'
+///
+/// /// document on machine A
+/// const localDoc = new YDoc()
+/// const localSV = encodeStateVector(localDoc)
+///
+/// // document on machine B
+/// const remoteDoc = new YDoc()
+/// const remoteDelta = encodeStateAsUpdateV2(remoteDoc, localSV)
+///
+/// applyUpdateV2(localDoc, remoteDelta)
+/// ```
+#[wasm_bindgen(js_name = applyUpdateV2)]
+pub fn apply_update_v2(doc: &mut YDoc, diff: Uint8Array) {
+    doc.begin_transaction().apply_v2(diff);
+}
+
 /// A transaction that serves as a proxy to document block store. Ywasm shared data types execute
 /// their operations in a context of a given transaction. Each document can have only one active
 /// transaction at the time - subsequent attempts will cause exception to be thrown.
@@ -448,6 +497,46 @@ impl YTransaction {
         Uint8Array::from(&payload[..payload.len()])
     }
 
+    /// Encodes all updates that have happened since a given version `vector` into a compact delta
+    /// representation using lib0 v1 encoding. If `vector` parameter has not been provided, generated
+    /// delta payload will contain all changes of a current ywasm document, working effectively as
+    /// its state snapshot.
+    ///
+    /// Example:
+    ///
+    /// ```javascript
+    /// import YDoc from 'ywasm'
+    ///
+    /// /// document on machine A
+    /// const localDoc = new YDoc()
+    /// const localTxn = localDoc.beginTransaction()
+    ///
+    /// // document on machine B
+    /// const remoteDoc = new YDoc()
+    /// const remoteTxn = localDoc.beginTransaction()
+    ///
+    /// try {
+    ///     const localSV = localTxn.stateVectorV1()
+    ///     const remoteDelta = remoteTxn.diffV2(localSv)
+    ///     localTxn.applyV2(remoteDelta)
+    /// } finally {
+    ///     localTxn.free()
+    ///     remoteTxn.free()
+    /// }
+    /// ```
+    #[wasm_bindgen(js_name = diffV2)]
+    pub fn diff_v2(&self, vector: Option<Uint8Array>) -> Uint8Array {
+        let mut encoder = EncoderV2::new();
+        let sv = if let Some(vector) = vector {
+            StateVector::decode_v1(vector.to_vec().as_slice())
+        } else {
+            StateVector::default()
+        };
+        self.0.encode_diff(&sv, &mut encoder);
+        let payload = encoder.to_vec();
+        Uint8Array::from(&payload[..payload.len()])
+    }
+
     /// Applies delta update generated by the remote document replica to a current transaction's
     /// document. This method assumes that a payload maintains lib0 v1 encoding format.
     ///
@@ -481,11 +570,50 @@ impl YTransaction {
         self.0.apply_update(update)
     }
 
+    /// Applies delta update generated by the remote document replica to a current transaction's
+    /// document. This method assumes that a payload maintains lib0 v2 encoding format.
+    ///
+    /// Example:
+    ///
+    /// ```javascript
+    /// import YDoc from 'ywasm'
+    ///
+    /// /// document on machine A
+    /// const localDoc = new YDoc()
+    /// const localTxn = localDoc.beginTransaction()
+    ///
+    /// // document on machine B
+    /// const remoteDoc = new YDoc()
+    /// const remoteTxn = localDoc.beginTransaction()
+    ///
+    /// try {
+    ///     const localSV = localTxn.stateVectorV1()
+    ///     const remoteDelta = remoteTxn.diffV2(localSv)
+    ///     localTxn.applyV2(remoteDelta)
+    /// } finally {
+    ///     localTxn.free()
+    ///     remoteTxn.free()
+    /// }
+    /// ```
+    #[wasm_bindgen(js_name = applyV2)]
+    pub fn apply_v2(&mut self, diff: Uint8Array) {
+        let diff: Vec<u8> = diff.to_vec();
+        let mut decoder = DecoderV2::from(diff.as_slice());
+        let update = Update::decode(&mut decoder);
+        self.0.apply_update(update)
+    }
+
     #[wasm_bindgen(js_name = encodeUpdate)]
     pub fn encode_update(&mut self) -> Uint8Array {
         let diff = self.0.encode_update_v1();
         Uint8Array::from(&diff[..diff.len()])
     }
+
+    #[wasm_bindgen(js_name = encodeUpdateV2)]
+    pub fn encode_update_v2(&mut self) -> Uint8Array {
+        let diff = self.0.encode_update_v2();
+        Uint8Array::from(&diff[..diff.len()])
+    }
 }
 
 /// Event generated by `YArray.observe` method. Emitted during transaction commit phase.
@@ -1053,6 +1181,15 @@ impl From<Subscription<XmlTextEvent>> for YXmlTextObserver {
     }
 }
 
+#[wasm_bindgen]
+pub struct YEventObserver(Subscription<Events>);
+
+impl From<Subscription<Events>> for YEventObserver {
+    fn from(o: Subscription<Events>) -> Self {
+        YEventObserver(o)
+    }
+}
+
 enum SharedType<T, P> {
     Integrated(T),
     Prelim(P),
@@ -1226,7 +1363,7 @@ impl YText {
                 let tuple = js_sys::Array::from(&tuple);
                 let key: String = tuple.get(0).as_string()?;
                 let value = js_into_any(&tuple.get(1))?;
-                map.insert(key.into_boxed_str(), value);
+                map.insert(key.into(), value);
             }
             Some(map)
         } else {
@@ -1272,7 +1409,7 @@ impl YText {
 
     /// Subscribes to all operations happening over this instance of `YText`. All changes are
     /// batched and eventually triggered during transaction commit phase.
-    /// Returns an `YObserver` which, when free'd, will unsubscribe current callback.
+    /// Returns an `YTextObserver` which, when free'd, will unsubscribe current callback.
     #[wasm_bindgen(js_name = observe)]
     pub fn observe(&mut self, f: js_sys::Function) -> YTextObserver {
         match &mut *self.0.borrow_mut() {
@@ -1288,6 +1425,25 @@ impl YText {
             }
         }
     }
+
+    /// Subscribes to all operations happening over this Y shared type, as well as events in
+    /// shared types stored within this one. All changes are batched and eventually triggered
+    /// during transaction commit phase.
+    /// Returns an `YEventObserver` which, when free'd, will unsubscribe current callback.
+    #[wasm_bindgen(js_name = observeDeep)]
+    pub fn observe_deep(&mut self, f: js_sys::Function) -> YEventObserver {
+        match &mut *self.0.borrow_mut() {
+            SharedType::Integrated(v) => v
+                .observe_deep(move |txn, e| {
+                    let arg = events_into_js(txn, e);
+                    f.call1(&JsValue::UNDEFINED, &arg).unwrap();
+                })
+                .into(),
+            SharedType::Prelim(_) => {
+                panic!("YText.observeDeep is not supported on preliminary type.")
+            }
+        }
+    }
 }
 
 /// A collection used to store data in an indexed sequence structure. This type is internally
@@ -1492,6 +1648,25 @@ impl YArray {
             }
         }
     }
+
+    /// Subscribes to all operations happening over this Y shared type, as well as events in
+    /// shared types stored within this one. All changes are batched and eventually triggered
+    /// during transaction commit phase.
+    /// Returns an `YEventObserver` which, when free'd, will unsubscribe current callback.
+    #[wasm_bindgen(js_name = observeDeep)]
+    pub fn observe_deep(&mut self, f: js_sys::Function) -> YEventObserver {
+        match &mut *self.0.borrow_mut() {
+            SharedType::Integrated(v) => v
+                .observe_deep(move |txn, e| {
+                    let arg = events_into_js(txn, e);
+                    f.call1(&JsValue::UNDEFINED, &arg).unwrap();
+                })
+                .into(),
+            SharedType::Prelim(_) => {
+                panic!("YText.observeDeep is not supported on preliminary type.")
+            }
+        }
+    }
 }
 
 fn to_iter(iterator: JsValue) -> JsValue {
@@ -1768,6 +1943,25 @@ impl YMap {
             }
         }
     }
+
+    /// Subscribes to all operations happening over this Y shared type, as well as events in
+    /// shared types stored within this one. All changes are batched and eventually triggered
+    /// during transaction commit phase.
+    /// Returns an `YEventObserver` which, when free'd, will unsubscribe current callback.
+    #[wasm_bindgen(js_name = observeDeep)]
+    pub fn observe_deep(&mut self, f: js_sys::Function) -> YEventObserver {
+        match &mut *self.0.borrow_mut() {
+            SharedType::Integrated(v) => v
+                .observe_deep(move |txn, e| {
+                    let arg = events_into_js(txn, e);
+                    f.call1(&JsValue::UNDEFINED, &arg).unwrap();
+                })
+                .into(),
+            SharedType::Prelim(_) => {
+                panic!("YText.observeDeep is not supported on preliminary type.")
+            }
+        }
+    }
 }
 
 #[wasm_bindgen]
@@ -2001,6 +2195,20 @@ impl YXmlElement {
             })
             .into()
     }
+
+    /// Subscribes to all operations happening over this Y shared type, as well as events in
+    /// shared types stored within this one. All changes are batched and eventually triggered
+    /// during transaction commit phase.
+    /// Returns an `YEventObserver` which, when free'd, will unsubscribe current callback.
+    #[wasm_bindgen(js_name = observeDeep)]
+    pub fn observe_deep(&mut self, f: js_sys::Function) -> YEventObserver {
+        self.0
+            .observe_deep(move |txn, e| {
+                let arg = events_into_js(txn, e);
+                f.call1(&JsValue::UNDEFINED, &arg).unwrap();
+            })
+            .into()
+    }
 }
 
 #[wasm_bindgen]
@@ -2233,6 +2441,20 @@ impl YXmlText {
             })
             .into()
     }
+
+    /// Subscribes to all operations happening over this Y shared type, as well as events in
+    /// shared types stored within this one. All changes are batched and eventually triggered
+    /// during transaction commit phase.
+    /// Returns an `YEventObserver` which, when free'd, will unsubscribe current callback.
+    #[wasm_bindgen(js_name = observeDeep)]
+    pub fn observe_deep(&mut self, f: js_sys::Function) -> YEventObserver {
+        self.0
+            .observe_deep(move |txn, e| {
+                let arg = events_into_js(txn, e);
+                f.call1(&JsValue::UNDEFINED, &arg).unwrap();
+            })
+            .into()
+    }
 }
 
 #[repr(transparent)]
@@ -2419,6 +2641,22 @@ fn xml_into_js(v: Xml) -> JsValue {
     }
 }
 
+fn events_into_js(txn: &Transaction, e: &Events) -> JsValue {
+    let mut array = js_sys::Array::new();
+    let mapped = e.iter().map(|e| {
+        let js: JsValue = match e {
+            Event::Text(e) => YTextEvent::new(e, txn).into(),
+            Event::Array(e) => YArrayEvent::new(e, txn).into(),
+            Event::Map(e) => YMapEvent::new(e, txn).into(),
+            Event::XmlElement(e) => YXmlEvent::new(e, txn).into(),
+            Event::XmlText(e) => YXmlTextEvent::new(e, txn).into(),
+        };
+        js
+    });
+    array.extend(mapped);
+    array.into()
+}
+
 enum Shared<'a> {
     Text(Ref<'a, YText>),
     Array(Ref<'a, YArray>),
